######################################################################
package BlueRoseChar;
######################################################################

=head1 NAME

BlueRoseChar -- object for reading/writing a Blue Rose RPG character

=head1 SYNOPSIS

    use BlueRoseChar;
    $char= new BlueRoseChar();
    $char= BlueRoseChar->getChar("shortname");
    $char->readFile($file);
    $char->writeFile(); # writes to $char->shortname().".txt"
    $char->writeFile($file);
    @statnames = $char->stats();
    $value = $char->$stat;          # return a stat value
    $char->$stat($newvalue);        # set a stat value
    $char->validate();

=head1 DESCRIPTION

This is an object-oriented module which stores the character 
information in a set of plain fields.  Each field uses only 
word characters, and corresponds to the JavaScript form 
element name, and the key in the parameter file.  

=cut

require 5.003;
use strict;
use Carp;
use CGI;
use FileHandle;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);

require Exporter;

@ISA     = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT  = qw( );
@EXPORT_OK = qw( );
$VERSION = '0.01';

######################################################################
# private variables
my @statdefs=qw(shortname 			word*25
		users 				string*40 
		CharacterName 			string*40 
		PlayerName 			string*40 
		CharacterRole			string*40 
		CharacterBackground		string*40 
		CharacterLevel			int*3[1:1..20]
		History 			text*2000
		Description 			text*1000
		Arcana 				text*1000
		Feats 				text*1000
		SpecialAbilities 		text*1000
		Equipment 			text*1000
		Conviction 			int*3[3:0..10]
		Corruption			int*3[0:0..10]
		Alignment			string*15
		Calling				string*40
		LightNature			string*40
		ShadowNature			string*40
		Wealth				int*3[0:0..20]
		Profession			string*30
		Reputation			int*3[0:0..20]
		Initiative 			int*3[0:-10..20]
		Defense 			int*3[0:-10..20]
		BaseAttackBonus 		int*3[0:-10..20]
		MeleeAttackBonus 		int*3[0:-10..20]
		RangedAttackBonus 		int*3[0:-10..20]
		Size 				string*10 
		Speed 				string*10
		AbilityPointsTotal 		int*3 
		Ability_Strength_Value 		int*3[0:-8..8]
		Ability_Dexterity_Value 	int*3[0:-8..8]
		Ability_Constitution_Value 	int*3[0:-8..8]
		Ability_Intelligence_Value 	int*3[0:-8..8]
		Ability_Wisdom_Value 		int*3[0:-8..8]
		Ability_Charisma_Value 		int*3[0:-8..8]
		SavingThrow_Toughness_Value 	int*3[0:-10..20]
		SavingThrow_Fortitude_Value 	int*3[0:-10..20]
		SavingThrow_Reflex_Value 	int*3[0:-10..20]
		SavingThrow_Will_Value 		int*3[0:-10..20]
		Skill_%_Name 			string*20
		Skill_%_Ability 		string*12
		Skill_%_Value 			int*3
		Skill_%_Modifier 		int*3[0:-10..10]
		Skill_%_Known 			int*1
		Skill_%_Favored 		int*1
		Weapon_%_Name 			string*20
		Weapon_%_AttackBonus 		int*3[0:-5..20]
		Weapon_%_DamageBonus 		int*3[0:-5..20]
		Weapon_%_Range 			string*10
		Weapon_%_AttackModifier 	int*3[0:-5..20]
		Weapon_%_DamageModifier 	int*3[0:-5..20]
		Weapon_%_Critical 		string*10
		Weapon_%_Weight 		string*10
		Weapon_%_WeaponType 		string*10
		Weapon_%_WeaponSize 		string*10
		Weapon_%_Hand 			int*1
		Weapon_%_Ranged 		int*1
		Weapon_%_IntelligenceBased 	int*1
		Weapon_%_WisdomBased 		int*1
		Weapon_%_CharismaBased 		int*1
		Weapon_%_SpecialNotes 		string*40
		);
my %statdef_of_stat = @statdefs;
my @statlist = grep { not m/\*/ } @statdefs;

######################################################################
my %wpnstats = 
    ('gauntlet'=>       [  4,  0,   'x2',    undef,    1,   'B','M',1,0],
     'unarmed'=>        [undef,0,   'x2',    undef,  undef, 'B','M',1,0],
     'hooves'=>         [undef,1,   'x2',    undef,  undef, 'B','M',1,0],
     'dagger'=>         [  4,  1,'19-20/x2', '10 ft',  1,  'SP','M',1,1],
     'spiked gauntlet'=>[  7,  1,   'x2',    undef,    1,   'P','M',1,0],
     'light mace'=>     [  7,  2,   'x2',    undef,    4,   'B','M',1,0],
     'sickle'=>         [  7,  2,   'x2',    undef,    2,   'S','M',1,0],
     'club'=>           [undef,2,   'x2',    '10 ft',  3,   'B','M',1,1],
     'heavy mace'=>     [  8,  3,   'x2',    undef,    8,   'B','M',1,0],
     'morningstar'=>    [ 10,  3,   'x2',    undef,    6,  'BP','M',1,0],
     'shortspear'=>     [  3,  2,   'x2',   '20 ft',   3,   'P','M',1,1],
     'longspear'=>      [  7,  3,   'x3',    undef,    9,   'P','M',1,0],
     'quarterstaff'=>   [undef,2,   'x2',    undef,    4,   'B','M',1,0],
     'spear'=>          [  4,  3,   'x3',   '20 ft',   6,   'P','M',1,1],
     'heavy crossbow'=> [  7,  4,'19-20/x2','120 ft',  8,   'P','M',0,1],
     'light crossbow'=> [  5,  3,'19-20/x2','80 ft',   4,   'P','M',0,1],
     'cryston'=>        [ 10,  0,   'x2',   'special', 2,   'F','M',0,1],
     'dart'=>           [  1,  1,   'x2',   '20 ft', 0.5,   'P','M',0,1],
     'javelin'=>        [  3,  2,   'x2',   '30 ft',   2,   'P','M',1,1],
     'sling'=>          [undef,1,   'x2',   '50 ft',   0,   'B','M',0,1],
     'throwing axe'=>   [  8,  2,   'x2',   '10 ft',   2,   'S','M',1,1],
     'light hammer'=>   [  3,  1,   'x2',   '20 ft',   2,   'B','M',1,1],
     'handaxe'=>        [  7,  2,   'x3',    undef,    3,   'S','M',1,0],
     'kukri'=>          [  7,  1,'18-20/x2', undef,    2,   'S','M',1,0],
     'light pick'=>     [  6,  1,   'x4',    undef,    3,   'P','M',1,0],
     'sap'=>            [  3,  2,   'x2',    undef,    2,   'B','M',1,0],
     'light shield'=>   [undef,0,   'x2',    undef,    0,   'B','M',1,0],
     'short sword'=>    [  9,  2,   'x2',    undef,    2,   'S','M',1,0],
     'battleaxe'=>      [  9,  3,   'x3',    undef,    6,   'S','M',1,0],
     'flail'=>          [  8,  3,   'x2',    undef,    5,   'B','M',1,0],
     'longsword'=>      [ 11,  3,'19-20/x2', undef,    4,   'S','M',1,0],
     'heavy pick'=>     [  8,  2,   'x4',    undef,    6,   'P','M',1,0],
     'rapier'=>         [ 12,  2,'18-20/x2', undef,    2,   'P','M',1,0],
     'scimitar'=>       [ 11,  2,'18-20/x2', undef,    4,   'S','M',1,0],
     'heavy shield'=>   [  8,  1,   'x2',    undef,   10,   'B','M',1,0],
     'trident'=>        [ 11,  3,   'x2',   '10 ft',   4,   'P','M',1,1],
     'warhammer'=>      [ 10,  3,   'x3',    undef,    5,   'B','M',1,0],
     'greataxe'=>       [ 12,  5,   'x3',    undef,   12,   'S','M',1,0],
     'greatclub'=>      [  6,  4,   'x2',    undef,    8,   'B','M',1,0],
     'heavy flail'=>    [ 11,  4,'19-20/x2', undef,   10,   'B','M',1,0],
     'greatsword'=>     [ 15,  4,'19-20/x2', undef,    8,   'S','M',1,0],
     'lance'=>          [  9,  3,   'x3',    undef,   10,   'P','M',1,0],
     'scythe'=>         [ 11,  2,   'x4',    undef,   10,   'S','M',1,0],
     'longbow'=>        [ 16,  3,   'x3',   '100 ft',  3,   'P','M',0,1],
     'composite longbow'=>[17, 3,   'x3',   '120 ft',  3,   'P','M',0,1],
     'shortbow'=>       [ 13,  2,   'x3',   '60 ft',   2,   'P','M',0,1],
     'composite shortbow'=>[16,2,   'x3',   '70 ft',   2,   'P','M',0,1],
     'bastard sword'=>  [ 14,  4,'19-20/x2', undef,    6,   'S','M',1,0],
     'whip'=>           [  3,  0,   'x2',    undef,    2,   'S','M',1,0],
     'bolas'=>          [  6,  1,   'x2',   '10 ft',   2,   'B','M',1,1],
     'net'=>            [ 12, undef, '-',   '10 ft',   6,    '','M',0,1]
     );

######################################################################
=head1 METHODS

The following methods are available:

=over 4

=item new (Constructor)
Create a new instance of this class by writing
    $mychar = BlueRoseChar->new();
=cut
sub new { 
    my $that= shift;  # either class name or template object
    my $self= {};     # $self is a hash reference
    bless($self, $that);
    return $self;
}

=item copy (Constructor)
Create a new instance of this class by writing
    $char2 = $char->copy();
=cut
sub copy { 
    my $self = shift; # either class name or template object
    my $copy= {};     # $self is a hash reference
    foreach (keys %$self) {
	$copy->{$_} = $self->{$_};
    }
    bless($copy, $self);
    return $copy;
}

=item getChar (Constructor)
Read a character object from a file by 
    $mychar = BlueRoseChar->getChar($shortname);
=cut
sub getChar { 
    my $self = new(shift);
    my $shortname = shift;
    # accept a filename as a shortname, converted by removing the ".txt"
    $shortname =~ s/\.txt$//;
    $self->readFile("$shortname.txt");
    return $self;
}

=item readFile
Read attributes and values into a character object from a file by 
    $mychar->readFile($filename);
=cut
sub readFile { 
    my $self= shift;
    my $filename = shift;
    my $fh = new FileHandle;
    unless (open($fh,$filename)) {
	carp "Problem opening file '$filename'"; 
	return $self;
    }
    my @keylist;
    while (<$fh>) {
	if (m/(\w+)(\{?)[\s:=]+(.*)/) {
	    my ($key,$startarray,$value)= ($1,$2,$3);
	    if ($startarray) {
		while (<$fh>) { last if (/^\s*\}/); $value .= $_; }
	    }
	    $self->{$key} = $value;
	    push(@keylist,$key);
	} elsif (m/<TEXT name=(\w+)>/i) {
	    my $key = $1;
	    my $value;
	    while (<$fh>) { 
		last if (/^\s*<\/TEXT>/i); 
		$value .= $_; 
	    }
	    $self->{$key} = $value;
	    push(@keylist,$key);
	} elsif (m/<TABLE type=(named|numbered) name=(\w+) columns=([\w\|]+)>/i) {
	    my ($type,$tablename,$columns) = ($1,$2,$3);
	    my @cols = split('|',$columns);
	    while (<$fh>) {
		last if (m/<\/TABLE>/i);
		my @vals = split(m/\s+/,$_);
		for my $i (0 .. $#cols) {
		    my $index = (lc($type) eq 'named') ? $vals[0] : ($i+1);
		    my $key = $tablename.'_'.$index.'_'.$cols[$i];
		    $self->{$key} = $vals[$i];
		    push(@keylist,$key);
		}
	    }
	}
    }    
    close($fh);
    $self->{"_BlueRoseChar_keylist"} = \@keylist;
    return $self;
}

=item writeFile
Write attributes and values from a character object to a file by 
    $mychar->writeFile($filename);
=cut
sub writeFile { 
    my $self= shift;
    my $filename = shift || ($self->shortname().".txt");
    my $fh = new FileHandle;
    if (not open($fh,">$filename")) {
	carp "File '$filename' was not found or could not be written to.";
	return $self;
    } else {
	foreach my $stat ($self->stats()) {
	    my $value = $self->$stat();
	    if ($value =~ m/\n/) { # multiline entry
		# remove leading and trailing whitespace including newlines
		$value =~ s/^\s*//s;  $value =~ s/\s*$//s;
		print $fh "<TEXT name=$stat>\n";
		print $fh "$value\n";
		print $fh "</TEXT>\n";
	    } else {
		print $fh "$stat : $value\n";
	    }
	} # foreach my $stat ($self->stats())
    } # end of if (not open(...)) { ... } else { ... }
    return $self;
}

sub statdef {
    my $self = shift(@_);
    my $stat = shift(@_);
    # replace numbered stat orderings with "%"
    $stat =~ s/_\d+_/_%_/;
    return $statdef_of_stat{$stat};
}

sub basestats {
    return @statlist;
}

sub stats {
    my $self = shift;
    croak "self is not a ref" unless (ref($self));
    # private variables start with an underscore
    my @selfstats = grep { not m|^_| } keys %{$self};
    my @sorted = sort { $self->statcmp($a,$b) } @selfstats;
    return @sorted;
}

sub abilities { 
    return grep { m/^Ability_.*_Value$/ } @statlist; 
}

sub skills { 
    return grep { m/^Skill_.*_Value$/ } @statlist; 
}

sub allskills { 
    my @allskills = ("Acrobatics","Bluff","Climb","Concentration","Craft",
		     "Diplomacy","Disable Device","Disguise","Escape Artist",
		     "Gather Information","Handle Animal","Heal","Intimidate",
		     "Jump","Knowledge","Language","Notice","Perform","Ride",
		     "Search","Sense Motive","Sleight of Hand","Sneak",
		     "Survival","Swim");
    return @allskills; 
}

sub text_of_key {
    my $self = shift(@_);
    my $key = shift(@_);
    # remove the type name of stat from the front
    $key =~ s/^(Ability_|Skill_)//;
    $key =~ s/^(Ability_|Skill_)//;
    # convert '_s_' to slash, '_c_' to colon, '_' to space
    $key =~ s/_s_/\//g; $key =~ s/_c_/\:/g; $key =~ s/_/ /g; 
    # add spaces before capital letters
    $key =~ s/([^ \/])([A-Z])/$1 $2/g;
    return $key;
}

sub statfield {
    my $self = shift(@_);
    my $req = shift(@_);
    my $name = shift(@_);
    my %opts = @_;
    # now see if there is a defined stat for this
    my $def = $self->statdef($name);
    # if Javascript is enabled, should pass the argument
    # -onChange=>"changeStat(this);"
    # to the textfield routine
    my $html;
    if (not defined($def)) {
	my $size = ($opts{'-size'} or 2);
	my $width = int(0.8*$size)+1;
	$html = $req->textfield(-name=>$name,-value=>$self->$name(),
				-size=>$width);
    } elsif ($def =~ m/^int\*(\d+)\[([\+\-]?\d+):([\+\-]?\d+)\.\.([\+\-]?\d+)\]/) {
	my ($size,$default,$min,$max) = ($1,$2,$3,$4);
	my $value = ($self->$name() or $default);
	$value =~ s/^\s*\+//;
	$html = $req->popup_menu(-name=>$name,-values=>[$min..$max],
				 -default=>$value);
    } elsif ($def =~ m/(string|int)\*(\d+)/) {
	my $size = ($opts{'-size'} or $2);
	my $width = int(0.8*$size)+1;
	$html = $req->textfield(-name=>$name,-value=>$self->$name(),-size=>$width);
    } elsif (defined($def) and $def =~ m/(text)\*(\d+)/) {
	my $size = ($opts{'-size'} or $2);
	my $cols = ($opts{'-cols'} or 60);
	my $rows = int($size/(5*$cols));
	$html = $req->textarea(-name=>$name,-value=>$self->$name(),
			       -rows=>$rows,-columns=>$cols);
    }
    return $html;
}

sub editfield {
    my $self = shift(@_);
    my $req = shift(@_);
    my $name= shift(@_);
    my $size= (shift(@_) || 2);
    my $html = statfield($name,$size);
    $html .= $req->button({-value=>"+",-onClick=>
			  "incrementStat(this.form.$name); return false;"});
    $html .= $req->button({-value=>"-",-onClick=>
                          "decrementStat(this.form.$name); return false;"});
    return $html;
}

sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) or croak "'$self' is not an object";
    my $name = $AUTOLOAD;
    $name =~ s/.*://; # strip fully-qualified portion of name
    croak "Bad method call '$name'" unless ($name =~ m/^\w/);
    my $def = $self->statdef($name);
    if (@_) {
	my $newval = shift;
	if ($self->{"_BlueRoseChar:printIfChange"} and ($newval ne $self->{$name})) {
	    print "Changing $name from $self->{$name} to $newval\n";
	}
	return $self->{$name} = $newval;
    } else {
	if (defined($def) and not defined($self->{$name})) {
	    return "";
	} else {
	    return $self->{$name};
	}
    }
}

######################################################################
sub validate {
    my $self = shift(@_);
    my %opt = @_;

    # derived stats
    if (defined($self->{'DefenseBonus'}) and not defined($self->{'Defense'})) {
	$self->{'Defense'} = $self->{'DefenseBonus'} + 10;
    } else {
	$self->{'DefenseBonus'} = $self->{'Defense'} - 10;
    }

    # process the skills
    $self->validateSkills(%opt);
    $self->validateWeapons(%opt);

    # calculate total ability points
    $self->{'AbilityPointsTotal'} = 0;
    foreach ($self->abilities()) {
	$self->{'AbilityPointsTotal'} += $self->{$_};
    }

    return $self;
} # end of sub validate

sub validateSkills {
    my $self = shift(@_);
    my %opt = @_;

    # process the skills
    foreach my $i (1 .. 30) {
	my $name = $self->{'Skill_'.$i.'_Name'};
	my $modifier = $self->{'Skill_'.$i.'_Modifier'};
	my $known = $self->{'Skill_'.$i.'_Known'};
	my $favored = $self->{'Skill_'.$i.'_Favored'};

	# calculate the value from abilities
	my $value = 0;
	$value = ($self->{'CharacterLevel'} || 1) + 3 if ($known);
	$value = int($value/2) if (not $favored);
	$value += $modifier if ($modifier =~ m/^\s*[\+\-]\d+\s*$/);
	my $ability = '';
	if ($name =~ m/^(Climb|Jump|Swim)/i) {
	    $ability = 'Strength';
	    $value += $self->{'Ability_Strength_Value'};
	} elsif ($name =~ m/^(Acrobatics|Escape Artist|Ride|Sleight of Hand|Sneak)/i) {
	    $ability = 'Dexterity';
	    $value += $self->{'Ability_Dexterity_Value'};
	} elsif ($name =~ m/^(Craft|Disable Device|Knowledge|Search)/i) {
	    $ability = 'Intelligence';
	    $value += $self->{'Ability_Intelligence_Value'};
	} elsif ($name =~ m/^(Concentration|Heal|Notice|Sense Motive|Survival)/i) {
	    $ability = 'Wisdom';
	    $value += $self->{'Ability_Wisdom_Value'};
	} elsif ($name =~ m/^(Bluff|Diplomacy|Disguise|Gather Information|Handle Animal|Intimidate|Perform)/i) {
	    $ability = 'Charisma';
	    $value += $self->{'Ability_Charisma_Value'};
	} else {
	    $ability = '';
	    $value = 0;
	}
	my $key1 = 'Skill_'.$i.'_Ability';
	$self->$key1($ability);
	my $key2 = 'Skill_'.$i.'_Value';
	$self->$key2($value);
    }

} # end of sub validateSkills

sub validateWeapons {
    my $self = shift(@_);
    my %opt = @_;

    # process the weapons
    foreach my $i (1 .. 10) {
	my $name = $self->{'Weapon_'.$i.'_Name'};
	# common fields
	my $attack = $self->{'Weapon_'.$i.'_AttackBonus'};
	my $damage = $self->{'Weapon_'.$i.'_DamageBonus'};
	my $range = $self->{'Weapon_'.$i.'_Range'};
	# optional fields
	my $cost = $self->{'Weapon_'.$i.'_Cost'};
	my $attackmod = $self->{'Weapon_'.$i.'_AttackModifier'} || 0;
	my $damagemod = $self->{'Weapon_'.$i.'_DamageModifier'} || 0;
	my $crit = $self->{'Weapon_'.$i.'_Critical'} || "x2";
	my $weight = $self->{'Weapon_'.$i.'_Weight'} || undef;
	my $type = $self->{'Weapon_'.$i.'_WeaponType'} || undef;
	my $size = $self->{'Weapon_'.$i.'_WeaponSize'} || 'M';
	my $ishand = $self->{'Weapon_'.$i.'_IsHand'} || undef;
	my $isranged = $self->{'Weapon_'.$i.'_IsRanged'} || undef;
	my $isint = $self->{'Weapon_'.$i.'_IntelligenceBased'} || undef;
	my $iswis = $self->{'Weapon_'.$i.'_WisdomBased'} || undef;
	my $ischa = $self->{'Weapon_'.$i.'_CharismaBased'} || undef;
	my $special = $self->{'Weapon_'.$i.'_SpecialNotes'} || undef;

	# try to match against the defined weapon keys
	my $key = lc($name);
	$key =~ s/^\s+//; $key =~ s/\s+$//; # trim whitespace
	$key =~ s/\s+/ /g; # reduce to single spaces
	$key =~ s/^(.*),\s*(\S.*)/$1 $2/; # 'mace, light' => 'light mace'

	foreach my $match (sort {length($b) <=> length($a)} keys %wpnstats) {
	    if ($key =~ m/$match/) {
		my $aref = $wpnstats{$match};
		$cost = $aref->[0] unless (defined($cost));
		$damagemod = $aref->[1] unless (defined($damagemod));
		$crit = $aref->[2] unless (defined($crit));
		$range = $aref->[3] unless (defined($range));
		$weight = $aref->[4] unless (defined($weight));
		$type = $aref->[5] unless (defined($type));
		$size = $aref->[6] unless (defined($size));
		$ishand = $aref->[7] unless (defined($ishand));
		$isranged = $aref->[8] unless (defined($isranged));
	    }
	}
	# Fill in either attack or attackmod
	if (defined($attack) and not defined($attackmod)) {
	    $attackmod = $attack;
	    $attackmod -= $self->{'BaseAttackBonus'};
	    if ($isranged) {
		$attackmod -= $self->{'Ability_Dexterity_Value'};
	    } elsif ($ishand) {
		$attackmod -= $self->{'Ability_Strength_Value'};
	    }
	} else {
	    $attack = $attackmod || 0;
	    $attack += $self->{'BaseAttackBonus'};
	    if ($isranged) {
		$attack += $self->{'Ability_Dexterity_Value'};
	    } elsif ($ishand) {
		$attack += $self->{'Ability_Strength_Value'};
	    }
	}
	# Fill in either damage or damagemod
	if (defined($damage) and not defined($damagemod)) {
	    $damagemod = $damage;
	    $damagemod -= $self->{'Ability_Strength_Value'} if ($ishand);
	} else {
	    $damage = $damagemod || 0;
	    $damage += $self->{'Ability_Strength_Value'} if ($ishand);
	}

	$self->{'Weapon_'.$i.'_AttackBonus'} = $attack;
	$self->{'Weapon_'.$i.'_AttackModifier'} = $attackmod;
	$self->{'Weapon_'.$i.'_DamageBonus'} = $damage;
	$self->{'Weapon_'.$i.'_DamageModifier'} = $damagemod;
	$self->{'Weapon_'.$i.'_Range'} = $range;
	$self->{'Weapon_'.$i.'_Critical'} = $crit;
	$self->{'Weapon_'.$i.'_Weight'} = $weight;
	$self->{'Weapon_'.$i.'_WeaponType'} = $type;
	$self->{'Weapon_'.$i.'_WeaponSize'} = $size;
	$self->{'Weapon_'.$i.'_IsHand'} = $ishand;
	$self->{'Weapon_'.$i.'_IsRanged'} = $isranged;
	$self->{'Weapon_'.$i.'_IntelligenceBased'} = $isint;
	$self->{'Weapon_'.$i.'_WisdomBased'} = $iswis;
	$self->{'Weapon_'.$i.'_CharismaBased'} = $ischa;
	$self->{'Weapon_'.$i.'_SpecialNotes'} = $special;
    } # end of foreach my $i (1 .. 10)

    return $self;
} # end of sub validateWeapons

sub shortenSkills {
    my ($self,$costflag,$printflag) = @_;
    my $j = 1;
    foreach my $i (1 .. 28) {
	my %fields;
	foreach my $f (qw(Name Modifier Known Favored Value)) {
	    $fields{$f} = $self->{'Skill_'.$i.'_'.$f};
	    delete($self->{'Skill_'.$i.'_'.$f});
	}
	if ($fields{'Known'} or $fields{'Favored'}) {
	    foreach my $f (qw(Name Modifier Known Favored Value)) {
		my $key = 'Skill_'.$j.'_'.$f;
		$self->$key($fields{$f});
	    }
	    $j++;
	}
    }
}
sub expandSkills {
    my ($self,$costflag,$printflag) = @_;
    my %hashref_of_name;
    my @crafts;
    my @knowledges;
    foreach my $i (1 .. 28) {
	my $name = $self->{'Skill_'.$i.'_Name'} or next;
	$hashref_of_name{$name} = {};
	foreach my $f (qw(Name Modifier Known Favored Value)) {
	    $hashref_of_name{$name}->{$f} = $self->{'Skill_'.$i.'_'.$f};
	    delete($self->{'Skill_'.$i.'_'.$f});
	}
	if (($name =~ m/Craft/) and (@crafts < 2)) {
	    push(@crafts,$name);
	} elsif (($name =~ m/Knowledge/) and (@knowledges < 3))  {
	    push(@knowledges,$name);
	}
    }
    while (scalar(@crafts) < 2) { push(@crafts,"Craft: "); }
    while (scalar(@knowledges) < 3) { push(@knowledges,"Knowledge: "); }
    my $i = 1;
    foreach my $skill ($self->allskills()) {
	if ($skill =~ m/^Craft/) {
	    # make sure we have at least two crafts
	    foreach my $name (@crafts) {
		$self->{'Skill_'.$i.'_Name'} = $name;
		foreach my $f (qw(Modifier Known Favored Value)) {
		    $self->{'Skill_'.$i.'_'.$f} = $hashref_of_name{$name}->{$f};
		}
		$i++;
	    }
	} elsif ($skill =~ m/^Knowledge/) {
	    # make sure we have at least three knowledges
	    foreach my $name (@knowledges) {
		$self->{'Skill_'.$i.'_Name'} = $name;
		foreach my $f (qw(Modifier Known Favored Value)) {
		    $self->{'Skill_'.$i.'_'.$f} = $hashref_of_name{$name}->{$f};
		}
		$i++;
	    }
	} else {
	    $self->{'Skill_'.$i.'_Name'} = $skill;
	    foreach my $f (qw(Modifier Known Favored Value)) {
		$self->{'Skill_'.$i.'_'.$f} = $hashref_of_name{$skill}->{$f};
	    }
	    $i++;
	} # end of if (craft) elsif (knowledge) else (...)
    } # foreach my $skill ($self->allskills())
}

######################################################################
sub statcmp {
    my ($self,$arga,$argb) = @_;
    my ($orda,$ordb) = ($self->statord($arga),$self->statord($argb));
    return ($orda <=> $ordb);
}
sub statord {
    my $self = shift(@_);
    my $arg = shift(@_);
    my $num = 0;
    if ($arg =~ s/_(\d+)_/_%_/) { $num = $1; }
    foreach my $i (0..$#statlist) {
	if ($arg eq $statlist[$i]) {
	    return 1000*1000 + 1000*$num + $i;
	}
    }
    return 2*1000*1000;
}

sub _max { return (sort {$b <=> $a} @_)[0]; }

sub _min { return (sort {$a <=> $b} @_)[0]; }

######################################################################

1; # required end-of-module line
